/**
 * Name........: SystemProperty
 * Description.: Listing of system properties used to be accessed.
 * Author......: Daniel Kasmeroglu
 * E-Mail......: daniel.kasmeroglu@kasisoft.net
 * Company.....: Kasisoft
 * License.....: LGPL
 */
package com.kasisoft.libs.common.constants;

import java.util.*;

import java.io.*;

/**
 * Listing of system properties used to be accessed.
 */
public enum SystemProperty {

  ClassPath             ( "java.class.path"             , "classpath"   , true  ),
  ClassVersion          ( "java.class.version"          , "classver"    , false ),
  EndorsedDirs          ( "java.endorsed.dirs"          , "endorsed"    , true  ),
  ExtDirs               ( "java.ext.dirs"               , "extdirs"     , true  ),
  FileEncoding          ( "file.encoding"               , "encoding"    , false ),
  FileSeparator         ( "file.separator"              , "filesep"     , true  ),
  JavaHome              ( "java.home"                   , "jre"         , true  ),
  JavaVersion           ( "java.version"                , "javaver"     , false ),
  LineSeparator         ( "line.separator"              , "linesep"     , false ),
  OsName                ( "os.name"                     , "os"          , false ),
  Path                  ( "java.library.path"           , "path"        , true  ),
  RuntimeVersion        ( "java.runtime.version"        , "runtimever"  , false ),
  SpecificationVersion  ( "java.specification.version"  , "specver"     , false ),
  TempDir               ( "java.io.tmpdir"              , "temp"        , true  ),
  UserDir               ( "user.dir"                    , "pwd"         , true  ),
  UserHome              ( "user.home"                   , "home"        , true  );
  
  private String     key;
  private String     shortkey;
  private boolean    filesystem;
  
  /**
   * Sets up this value with the associated property key.
   * 
   * @param propertykey     The property key to be used. Neither <code>null</code> nor empty.
   * @param propertyshort   A short key mainly used for substitution purposes. Neither <code>null</code> nor empty.
   * @param filesys         <code>true</code> <=> The property is related to a filesystem value.
   */
  SystemProperty( String propertykey, String propertyshort, boolean filesys ) {
    key         = propertykey;
    shortkey    = propertyshort;
    filesystem  = filesys;
  }
  
  /**
   * Returns the property key allowing to access a system property.
   * 
   * @return   The property key allowing to access a system property. Neither <code>null</code> nor empty.
   */
  public String getKey() {
    return key;
  }
  
  /**
   * Returns the short key allowing to access a system property.
   * 
   * @return   The short key allowing to access a system property. Neither <code>null</code> nor empty.
   */
  public String getShortKey() {
    return shortkey;
  }
  
  /**
   * Returns the value for this property from the system properties.
   * 
   * @return   The value associated with the property.
   */
  public String getValue() {
    return getValue( System.getProperties(), false );
  }
  
  /**
   * Returns the value for this property. If the property cannot be found in the supplied map the
   * system properties will be used.
   * 
   * @param properties   The properties used to access the value.
   * 
   * @return   The value associated with the property.
   */
  public String getValue( Properties properties ) {
    return getValue( properties, true );
  }
  
  /**
   * Returns the value for this property.
   * 
   * @param properties   The properties used to access the value.
   * @param fallback     <code>true</code> <=> Use system properties as a fallback solution.
   * 
   * @return   The value associated with the property.
   */
  public String getValue( Properties properties, boolean fallback ) {
    String defvalue = null;
    if( fallback ) {
      defvalue = System.getProperty( getKey() );
    }
    String result = properties.getProperty( getKey(), defvalue );
    return processValue( result );
  }

  /**
   * Returns the value for this property. If the property cannot be found in the supplied map the system properties will 
   * be used.
   * 
   * @param properties   The properties used to access the value.
   * 
   * @return   The value associated with the property.
   */
  public String getValue( Map<String,String> properties ) {
    return getValue( properties, true );
  }
  
  /**
   * Returns the value for this property.
   * 
   * @param properties   The properties used to access the value.
   * @param fallback     <code>true</code> <=> Use system properties as a fallback solution.
   * 
   * @return   The value associated with the property.
   */
  public String getValue( Map<String,String> properties, boolean fallback ) {
    String defvalue = null;
    if( fallback ) {
      defvalue = System.getProperty( getKey() );
    }
    String result   = defvalue;
    if( properties.containsKey( getKey() ) ) {
      result = properties.get( getKey() );
    }
    return processValue( result );
  }

  /**
   * Processes the supplied value if necessary.
   * 
   * @param value   The value which has to be altered. Maybe <code>null</code>.
   * 
   * @return   The altered value. Maybe <code>null</code>.
   */
  private String processValue( String value ) {
    if( value != null ) {
      if( filesystem ) {
        // just make sure since the properties might come from an external map,
        // so we should deliver valid pathes
        value = value.replace( '\\', '/' ).replace( '/', File.separatorChar );
      }
    }
    return value;
  }
  
  /**
   * {@inheritDoc}
   */
  @Override
  public String toString() {
    String result = getValue();
    if( result == null ) {
      result = String.format( "%s()", name() );
    }
    return result;
  }
  
  /**
   * Returns the enumeration value associated with the supplied short key.
   * 
   * @param key   The short key used to refer to a system property. Neither <code>null</code> nor empty.
   *              
   * @return   The enumeration value or <code>null</code> in case the key could not be found.
   */
  public static SystemProperty getByShortkey( String key ) {
    for( SystemProperty sysprop : SystemProperty.values() ) {
      if( key.equals( sysprop.getShortKey() ) ) {
        return sysprop;
      }
    }
    return null;
  }

  /**
   * Returns the enumeration value associated with the supplied long key.
   * 
   * @param key   The long key used to refer to a system property. Neither <code>null</code> nor empty.
   *              
   * @return   The enumeration value or <code>null</code> in case the key could not be found.
   */
  public static SystemProperty getByLongkey( String key ) {
    for( SystemProperty sysprop : SystemProperty.values() ) {
      if( key.equals( sysprop.getKey() ) ) {
        return sysprop;
      }
    }
    return null;
  }

  /**
   * Returns the enumeration value associated with the supplied key (either the long or the short one).
   * 
   * @param key   The short/long key used to refer to a system property. Neither <code>null</code> nor empty.
   *              
   * @return   The enumeration value or <code>null</code> in case the key could not be found.
   */
  public static SystemProperty getByKey( String key ) {
    for( SystemProperty sysprop : SystemProperty.values() ) {
      if( key.equals( sysprop.getKey() ) || key.equals( sysprop.getShortKey() ) ) {
        return sysprop;
      }
    }
    return null;
  }

  /**
   * Creates a replacement map used to substitute system properties. The key is encapsulated by '%' characters.
   * 
   * @param shortkey     <code>true</code> <=> Use the short key representation to create the key.
   *
   * @return   A Map containing key-value pairs for a possible replacement. Not <code>null</code>.
   */
  public static Map<String,String> createReplacementMap( boolean shortkey ) {
    return createReplacementMap( "%%%s%%", shortkey );
  }
  
  /**
   * Creates a replacement map used to substitute system properties.
   * 
   * @param format     A formatting String with one %s format code. This is used in order to support various key 
   *                   formats. Neither <code>null</code> nor empty.
   * @param shortkey   <code>true</code> <=> Use the short key representation to create the key.
   *
   * @return   A Map containing key-value pairs for a possible replacement. Not <code>null</code>.
   */
  public static Map<String,String> createReplacementMap( String format, boolean shortkey ) {
    Map<String,String> result = new Hashtable<String,String>();
    for( SystemProperty sysprop : SystemProperty.values() ) {
      if( shortkey ) {
        result.put( String.format( format, sysprop.getShortKey() ), sysprop.getValue() );
      } else {
        result.put( String.format( format, sysprop.getKey() ), sysprop.getValue() );
      }
    }
    return result;
  }
  
} /* ENDENUM */
